from ..driver import DBDriver
import sqlalchemy
from sqlalchemy_privileges import GrantPrivileges  # type: ignore[import-untyped]
from sqlalchemy.schema import Table, MetaData
from records_mover.records import RecordsSchema
from records_mover.records.records_format import BaseRecordsFormat, AvroRecordsFormat
import logging
from ...utils.limits import (INT16_MIN, INT16_MAX,
                             INT32_MIN, INT32_MAX,
                             INT64_MIN, INT64_MAX,
                             FLOAT32_SIGNIFICAND_BITS,
                             FLOAT64_SIGNIFICAND_BITS,
                             num_digits)
from .sql import schema_sql_from_admin_views
import timeout_decorator
from typing import Optional, Union, Dict, List, Tuple
from ...url.base import BaseDirectoryUrl
from .unloader import RedshiftUnloader
from ..unloader import Unloader
from .loader import RedshiftLoader
from ..loader import LoaderFromRecordsDirectory
from ...check_db_conn_engine import check_db_conn_engine


logger = logging.getLogger(__name__)


class RedshiftDBDriver(DBDriver):
    def __init__(self,
                 db: Optional[Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection]],
                 db_conn: Optional[sqlalchemy.engine.Connection] = None,
                 db_engine: Optional[sqlalchemy.engine.Engine] = None,
                 s3_temp_base_loc: Optional[BaseDirectoryUrl] = None,
                 **kwargs) -> None:
        super().__init__(db=db, db_conn=db_conn, db_engine=db_engine)
        self.s3_temp_base_loc = s3_temp_base_loc
        self._redshift_loader =\
            RedshiftLoader(db=db,
                           db_conn=db_conn,
                           db_engine=db_engine,
                           meta=self.meta,
                           s3_temp_base_loc=s3_temp_base_loc)
        self._redshift_unloader =\
            RedshiftUnloader(db=db,
                             db_conn=db_conn,
                             db_engine=db_engine,
                             table=self.table,
                             s3_temp_base_loc=s3_temp_base_loc)

    def schema_sql(self, schema: str, table: str) -> str:
        out = schema_sql_from_admin_views(schema, table, None, db_conn=self.db_conn)
        if out is None:
            return super().schema_sql(schema=schema, table=table)
        else:
            return out

    # if this timeout goes off (at least for Redshift), it's probably
    # because memory is filling because sqlalchemy's cache of all
    # tables and columns filled up memory in the job.
    @timeout_decorator.timeout(80)
    def table(self, schema: str, table: str) -> Table:
        with self.db_engine.connect() as conn:
            with conn.begin():
                # The code in the Redshift SQLAlchemy driver relies on 'SET
                # LOCAL search_path' in order to do reflection and pull back
                # column information on tables.
                #
                # Unfortunately, 'SET LOCAL' only works when a transaction is
                # in process, and it looks like there's often not here, and an
                # exception will be thrown unless the table you're working
                # with is in your default schema.
                #
                # There was some hang-wringing when they added it originally:
                #  https://github.com/sqlalchemy-redshift/sqlalchemy-redshift/commit/
                #      94723ec6437c5e5197fcf785845499e81640b167
                return Table(table, self.meta, schema=schema, autoload_with=conn)

    def set_grant_permissions_for_groups(self,
                                         schema_name: str,
                                         table: str,
                                         groups: Dict[str, List[str]],
                                         db: Optional[Union[sqlalchemy.engine.Engine,
                                                      sqlalchemy.engine.Connection]],
                                         db_conn: Optional[sqlalchemy.engine.Connection] = None,
                                         db_engine: Optional[sqlalchemy.engine.Engine] = None
                                         ) -> None:
        db, db_conn, db_engine = check_db_conn_engine(db=db, db_conn=db_conn, db_engine=db_engine)
        for perm_type in groups:
            groups_list = groups[perm_type]
            for group in groups_list:
                if not perm_type.isalpha():
                    raise TypeError("Please make sure your permission types"
                                    " are an acceptable value.")
                table_obj = Table(table, MetaData(), schema=schema_name)
                perms_sql = str(GrantPrivileges(perm_type, table_obj, group))
                if db_conn:
                    db_conn.execute(perms_sql)
                else:
                    with db_engine.connect() as conn:
                        conn.execute(perms_sql)
        return None

    def supports_time_type(self) -> bool:
        return False

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.SMALLINT):
            return (INT16_MIN, INT16_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.INTEGER):
            return (INT32_MIN, INT32_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.BIGINT):
            return (INT64_MIN, INT64_MAX)
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.REAL):
            return (32, FLOAT32_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT16_MIN and max_value <= INT16_MAX:
                return sqlalchemy.sql.sqltypes.SMALLINT()
            if min_value >= INT32_MIN and max_value <= INT32_MAX:
                return sqlalchemy.sql.sqltypes.INTEGER()
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                return sqlalchemy.sql.sqltypes.BIGINT()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if precision > 38:
            logger.warning("Using floating point type, as Redshift can't store >38 "
                           f"digits of precision in a fixed point type (requested: {precision})")
            return sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION()
        return super().type_for_fixed_point(precision=precision,
                                            scale=scale)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning(f"Falling back to Redshift DOUBLE PRECISION type, as Redshift "
                           "doesn't support fp_significand_bits>{FLOAT64_SIGNIFICAND_BITS} "
                           f"(requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._redshift_loader

    def loader_from_fileobj(self) -> None:
        return None

    def unloader(self) -> Optional[Unloader]:
        return self._redshift_unloader

    def tweak_records_schema_for_load(self,
                                      records_schema: RecordsSchema,
                                      records_format: BaseRecordsFormat) -> RecordsSchema:
        if isinstance(records_format, AvroRecordsFormat):
            # upon testing, Redshift does not seem to support any of
            # Avro's logicalTypes - here's an example Avro schema; the
            # data will only import into the base type, not the
            # logical type :
            #
            # {"name": "date", "type": ["null", {"type": "int",
            #                                    "logicalType": "date" }]},
            # {"name": "time", "type": [ "null", {"type": "long",
            #                                     "logicalType": "time-micros"}]},
            # {"name": "timestamp", "type": [ "null", {"type": "string",
            #                                          "logicalType": "datetime"}]},
            # {"name": "timestamptz", "type": [ "null", {"type": "long",
            #                                            "logicalType": "timestamp-micros"}]}]}
            #
            # We could potentially tell Redshift to use
            # TIMEFORMAT/DATEFORMAT on import.  Unfortunately, while
            # it supports 'epochsecs' and 'epochmillis', it doesn't
            # support 'epochdays', or 'epochmicros', which would be
            # needed for the above.
            return records_schema.cast_field_types({
                'date': 'integer',
                'time': 'integer',
                'datetime': 'string',
                # timestamp-micros is too large for default integer size
                'datetimetz': 'string',
            })
        else:
            return records_schema



# flake8: noqa

import sqlalchemy
from typing import Union, Optional, Tuple
import warnings


def check_db_conn_engine(db: Optional[Union[sqlalchemy.engine.Engine,
                                            sqlalchemy.engine.Connection]] = None,
                         db_conn: Optional[sqlalchemy.engine.Connection] = None,
                         db_engine: Optional[sqlalchemy.engine.Engine] = None) -> \
    Tuple[Optional[Union[sqlalchemy.engine.Engine,
                         sqlalchemy.engine.Connection]],
          Optional[sqlalchemy.engine.Connection],
          sqlalchemy.engine.Engine]:
    """
    This function was added because, in previous versions of SQL Alchemy, you could execute SQL against either
    Engines or Connections and not worry very much about the difference between the two.  Since version 2, this
    is no longer the case.

    Because of the interchangeability between Engines and Connections, most of our code accepted a rather ambiguous
    'db' argument, which could be either an Engine or a Connection.  This function is intended to be used to disabiguate
    the two, create engines from connections where possible, and warn users to stop using the 'db' argument.

    Args:
        db: Deprecated argument. Use db_conn for Connection objects and db_engine for Engine objects.
        db_conn: Optional SQLAlchemy Connection object.
        db_engine: Optional SQLAlchemy Engine object.

    Returns:
        A tuple containing the Connection or Engine 'db' object provided as input, the Connection object if provided,
        and the Engine object.

    Raises:
        ValueError: If none of db, db_conn, or db_engine are provided as arguments.
    """
    if db:
        warnings.warn("The db argument is deprecated and will be"
                      "removed in future releases.\n"
                      "Please use db_conn for Connection objects and db_engine for Engine"
                      "objects.",
                      DeprecationWarning)
    if not (db or db_conn or db_engine):
        raise ValueError("Either db, db_conn, or db_engine must be provided as arguments")
    if isinstance(db, sqlalchemy.engine.Connection) and not db_conn:
        db_conn = db
    if isinstance(db, sqlalchemy.engine.Engine) and not db_engine:
        db_engine = db
    if not db_engine:
        print("db_engine is not provided, so we're assigning db_engine to db_conn.engine")
        db_engine = db_conn.engine  # type: ignore[union-attr]
    return (db, db_conn, db_engine)  # type: ignore[return-value]



from typing import Optional
from .base import (SupportsMoveToRecordsDirectory,
                   SupportsToDataframesSource)
from ...db.quoting import quote_schema_and_table
from ...db import DBDriver
from ..records_directory import RecordsDirectory
from ..processing_instructions import ProcessingInstructions
from ..records_format import BaseRecordsFormat
from ..unload_plan import RecordsUnloadPlan
from ..results import MoveResult
from sqlalchemy import Table, MetaData, select
from sqlalchemy.engine import Engine
from contextlib import contextmanager
from ..schema import RecordsSchema
from ...url.resolver import UrlResolver
from records_mover.url.base import BaseDirectoryUrl
import logging
from typing import Generator, Iterator, List, TYPE_CHECKING
if TYPE_CHECKING:
    from .dataframes import DataframesRecordsSource  # noqa
    from pandas import DataFrame  # noqa

logger = logging.getLogger(__name__)


class TableRecordsSource(SupportsMoveToRecordsDirectory,
                         SupportsToDataframesSource):
    records_format: Optional[BaseRecordsFormat]

    def __init__(self,
                 schema_name: str,
                 table_name: str,
                 driver: DBDriver,
                 url_resolver: UrlResolver) -> None:
        self.schema_name = schema_name
        self.table_name = table_name
        self.driver = driver
        self.unloader = driver.unloader()
        if self.unloader is not None:
            self.records_format = self.unloader.best_records_format()
        else:
            self.records_format = None
        self.url_resolver = url_resolver

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        unloader = self.driver.unloader()
        if unloader is None:
            return []
        return unloader.known_supported_records_formats_for_unload()

    def can_move_to_format(self,
                           target_records_format: BaseRecordsFormat) -> bool:
        unloader = self.driver.unloader()
        if unloader is None:
            return False
        return unloader.can_unload_format(target_records_format)

    def can_move_to_scheme(self, scheme: str) -> bool:
        unloader = self.driver.unloader()
        if unloader is None:
            # bulk export is not provided by this database
            logger.warning("No unloader configured for this database "
                           f"type ({self.driver.db_engine.name})")
            return False
        can_unload = unloader.can_unload_to_scheme(scheme)
        if not can_unload:
            logger.warning(f"This database ({self.driver.db_engine.name}) is "
                           f"not configured to export to {scheme}")
        return can_unload

    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) -> \
            Iterator['DataframesRecordsSource']:
        from .dataframes import DataframesRecordsSource  # noqa
        import pandas

        db_conn = self.driver.db_conn
        db_engine = self.driver.db_engine
        records_schema = self.pull_records_schema()

        columns = db_engine.dialect.get_columns(db_conn, self.table_name, schema=self.schema_name)

        num_columns = len(columns)
        if num_columns == 0:
            raise LookupError(f"Could not find {self.schema_name}.{self.table_name}")
        entries_per_chunk = 2000000
        chunksize = int(entries_per_chunk / num_columns)
        logger.info(f"Exporting in chunks of up to {chunksize} rows by {num_columns} columns")

        meta = MetaData()
        table = Table(self.table_name, meta, schema=self.schema_name)
        quoted_table = quote_schema_and_table(None, self.schema_name,
                                              self.table_name, db_engine=db_engine,)
        logger.info(f"Reading {quoted_table}...")
        chunks: Generator['DataFrame', None, None] = \
            pandas.read_sql(select('*', table),  # type: ignore[arg-type]  # noqa: F821
                            con=db_conn,
                            chunksize=chunksize)
        try:
            yield DataframesRecordsSource(dfs=self.with_cast_dataframe_types(records_schema,
                                                                             chunks),
                                          records_schema=records_schema,
                                          processing_instructions=processing_instructions)
        finally:
            chunks.close()

    def with_cast_dataframe_types(self,
                                  records_schema: RecordsSchema,
                                  dfs: Iterator['DataFrame']) -> Iterator['DataFrame']:
        for df in dfs:
            yield records_schema.cast_dataframe_types(df)

    def pull_records_schema(self) -> RecordsSchema:
        return RecordsSchema.from_db_table(self.schema_name, self.table_name,
                                           driver=self.driver)

    def move_to_records_directory(self,
                                  records_directory: RecordsDirectory,
                                  records_format: BaseRecordsFormat,
                                  processing_instructions: ProcessingInstructions,
                                  schema_name: Optional[str] = None,
                                  table_name: Optional[str] = None,
                                  engine: Optional[Engine] = None) -> MoveResult:
        unload_plan = RecordsUnloadPlan(records_format=records_format,
                                        processing_instructions=processing_instructions)
        unloader = self.driver.unloader()
        if unloader is None:
            raise ValueError('This DBDriver does not support bulk unloading')
        export_count = unloader.unload(schema=self.schema_name, table=self.table_name,
                                       unload_plan=unload_plan,
                                       directory=records_directory)
        records_schema = self.pull_records_schema()
        records_directory.save_format(unload_plan.records_format)
        records_schema = self.driver.tweak_records_schema_after_unload(records_schema,
                                                                       unload_plan.records_format)
        records_directory.save_schema(records_schema)
        records_directory.finalize_manifest()

        return MoveResult(move_count=export_count, output_urls=None)

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        """Yield a temporary directory that can be used to call move_to_records_directory() on."""
        unloader = self.driver.unloader()
        if unloader is None:
            raise ValueError('This DBDriver does not support bulk unloading')
        with unloader.temporary_unloadable_directory_loc() as temp_loc:
            yield temp_loc

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.driver.db_engine.name})"



from typing import Callable, Optional, Type
from records_mover.records.results import MoveResult
from records_mover.db import DBDriver
from records_mover.records.prep import TablePrep
from records_mover.records.table import TargetTableDetails
from records_mover.records.existing_table_handling import ExistingTableHandling
import logging

logger = logging.getLogger(__name__)


def prep_and_load(tbl: TargetTableDetails,
                  prep: TablePrep,
                  schema_sql: str,
                  load: Callable[[DBDriver], Optional[int]],
                  load_exception_type: Type[Exception],
                  reset_before_reload: Callable[[], None] = lambda: None) -> MoveResult:
    logger.info("Connecting to database...")
    with tbl.db_engine.connect() as db_conn:
        with db_conn.begin():
            driver = tbl.db_driver(db=None, db_conn=db_conn)
            prep.prep(schema_sql=schema_sql, driver=driver)
    with tbl.db_engine.connect() as db_conn:
        with db_conn.begin():
            # This second transaction ensures the table has been created
            # before non-transactional statements like Redshift's COPY
            # take place.  Otherwise you'll get an error like:
            #
            #  Cannot COPY into nonexistent table
            driver = tbl.db_driver(db=None, db_conn=db_conn)
            try:
                import_count = load(driver)
            except load_exception_type:
                if not tbl.drop_and_recreate_on_load_error:
                    raise
                reset_before_reload()
                with tbl.db_engine.connect() as db_conn:
                    with db_conn.begin():
                        driver = tbl.db_driver(db=None, db_conn=db_conn)
                        prep.prep(schema_sql=schema_sql,
                                  driver=driver,
                                  existing_table_handling=ExistingTableHandling.DROP_AND_RECREATE)
                        import_count = load(driver)
            return MoveResult(move_count=import_count, output_urls=None)



import logging
import sqlalchemy
from typing import Optional

logger = logging.getLogger(__name__)


class DBConnMixin():
    _db_conn: Optional[sqlalchemy.engine.Connection]
    conn_opened_here: bool
    db_conn: sqlalchemy.engine.Connection
    db_engine: sqlalchemy.engine.Engine

    def get_db_conn(self) -> sqlalchemy.engine.Connection:
        if self._db_conn is None:
            self._db_conn = self.db_engine.connect()
            self.conn_opened_here = True
            logger.debug(f"Opened connection to database within {self} because none was provided.")
        return self._db_conn

    def set_db_conn(self,
                    db_conn: Optional[sqlalchemy.engine.Connection]) -> None:
        self._db_conn = db_conn

    def del_db_conn(self) -> None:
        if self.conn_opened_here:
            self.db_conn.close()
            self.db_conn = None  # type: ignore[assignment]  # noqa: F821
            self.conn_opened_here = False

    db_conn = property(get_db_conn, set_db_conn, del_db_conn)  # type: ignore[assignment]  # noqa: F821,E501



# flake8: noqa

from .driver import GenericDBDriver, DBDriver
import sqlalchemy
from typing import Union, Optional
from ..check_db_conn_engine import check_db_conn_engine


def db_driver(db: Optional[Union[sqlalchemy.engine.Engine,
                           sqlalchemy.engine.Connection]],
              db_conn: Optional[sqlalchemy.engine.Connection] = None,
              db_engine: Optional[sqlalchemy.engine.Engine] = None,
              **kwargs) -> DBDriver:
    db, db_conn, db_engine = check_db_conn_engine(db=db, db_conn=db_conn, db_engine=db_engine)
    engine_name: str = db_engine.name

    if engine_name == 'vertica':
        from .vertica.vertica_db_driver import VerticaDBDriver

        return VerticaDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)
    elif engine_name == 'redshift':
        from .redshift.redshift_db_driver import RedshiftDBDriver

        return RedshiftDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)
    elif engine_name == 'bigquery':
        from .bigquery.bigquery_db_driver import BigQueryDBDriver

        return BigQueryDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)
    elif engine_name == 'postgresql':
        from .postgres.postgres_db_driver import PostgresDBDriver

        return PostgresDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)
    elif engine_name == 'mysql':
        from .mysql.mysql_db_driver import MySQLDBDriver

        return MySQLDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)
    else:
        return GenericDBDriver(db=db, db_conn=db_conn, db_engine=db_engine, **kwargs)



from ...utils import quiet_remove
from ...records.delimited import cant_handle_hint
from records_mover.records.records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, AvroRecordsFormat
)
from records_mover.records.delimited import ValidatedRecordsHints
from records_mover.mover_types import _assert_never
from sqlalchemy_redshift.commands import Format, Encoding, Compression
from typing import Dict, Optional, Set

RedshiftCopyOptions = Dict[str, Optional[object]]


class RedshiftDelimitedRecordsHandler:
    """
    Class for handling Redshift delimited records format options.
    """

    def __init__(self,
                 records_format: DelimitedRecordsFormat,
                 fail_if_cant_handle_hint: bool,
                 unhandled_hints: Set[str],
                 fail_if_row_invalid: bool,
                 max_failure_rows: Optional[int]):
        """
        Initialize the RedshiftDelimitedRecordsHandler object.

        :param records_format: Delimited records format object.
        :param fail_if_cant_handle_hint: If True, raise an error if a hint is not handled.
        :param unhandled_hints: Set of unhandled hints.
        :param fail_if_row_invalid: If True, fail if a row is invalid.
        :param max_failure_rows: Maximum number of allowed failure rows.
        """
        self.redshift_options: RedshiftCopyOptions = dict()
        self.hints: ValidatedRecordsHints = records_format.\
            validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)
        self.unhandled_hints = unhandled_hints
        self.fail_if_cant_handle_hint = fail_if_cant_handle_hint
        self.fail_if_row_invalid = fail_if_row_invalid
        self.max_failure_rows = max_failure_rows
        self.process_compression()
        self.process_dateformat()
        self.process_encoding()
        self.process_quotechar()
        self.process_quoting()
        self.process_temporal_info()
        self.process_max_failure_rows()
        self.process_records_terminator()
        self.process_header_row()

    def process_compression(self):
        """
        Processes the compression hint for RedshiftCopyOptions.
        """
        compression_map = {
            'GZIP': Compression.gzip,
            'LZO': Compression.lzop,
            'BZIP': Compression.bzip2,
            None: None
        }
        compression = compression_map.get(self.hints.compression)
        if compression is None and self.hints.compression is not None:
            _assert_never(self.hints.compression)
        else:
            self.redshift_options['compression'] = compression
        quiet_remove(self.unhandled_hints, 'compression')

    def process_dateformat(self):
        """
        Processes the date format hint for RedshiftCopyOptions.
        """
        if self.hints.dateformat is None:
            self.redshift_options['date_format'] = 'auto'
        else:
            self.redshift_options['date_format'] = (self
                                                    .hints
                                                    .dateformat)
        quiet_remove(self.unhandled_hints, 'dateformat')

    def process_encoding(self):
        """
        Processes the encoding hint for RedshiftCopyOptions.
        """
        encoding_map = {
            'UTF8': Encoding.utf8,
            'UTF16': Encoding.utf16,
            'UTF16LE': Encoding.utf16le,
            'UTF16BE': Encoding.utf16be,
        }
        encoding = encoding_map.get(self.hints.encoding)
        if not encoding:
            cant_handle_hint(self.fail_if_cant_handle_hint,
                             'encoding', self.hints)
            self.redshift_options['encoding'] = Encoding(
                self.hints.encoding)
        else:
            self.redshift_options['encoding'] = encoding
        quiet_remove(self.unhandled_hints, 'encoding')

    def process_quotechar(self):
        """
        Processes the quote character hint for RedshiftCopyOptions.
        """
        self.redshift_options['quote'] = self.hints.quotechar
        quiet_remove(self.unhandled_hints, 'quotechar')

    def process_quoting(self):
        """
        Processes the quoting hint for RedshiftCopyOptions.
        """
        if self.hints.quoting == 'minimal':
            if self.hints.escape is not None:
                cant_handle_hint(self.fail_if_cant_handle_hint,
                                 'escape', self.hints)
            if self.hints.field_delimiter != ',':
                cant_handle_hint(self.fail_if_cant_handle_hint,
                                 'field-delimiter', self.hints)
            if self.hints.doublequote is not True:
                cant_handle_hint(self.fail_if_cant_handle_hint,
                                 'doublequote', self.hints)

            self.redshift_options['format'] = Format.csv
        else:
            self.redshift_options['delimiter'] = (self
                                                  .hints
                                                  .field_delimiter)
            if self.hints.escape == '\\':
                self.redshift_options['escape'] = True
            elif self.hints.escape is None:
                self.redshift_options['escape'] = False
            else:
                _assert_never(self.hints.escape)
            if self.hints.quoting == 'all':
                self.redshift_options['remove_quotes'] = True
                if self.hints.doublequote is not False:
                    cant_handle_hint(self.fail_if_cant_handle_hint,
                                     'doublequote', self.hints)

            elif self.hints.quoting is None:
                self.redshift_options['remove_quotes'] = False
            else:
                cant_handle_hint(self.fail_if_cant_handle_hint,
                                 'quoting', self.hints)
        quiet_remove(self.unhandled_hints, 'quoting')
        quiet_remove(self.unhandled_hints, 'escape')
        quiet_remove(self.unhandled_hints, 'field-delimiter')
        quiet_remove(self.unhandled_hints, 'doublequote')

    def process_temporal_info(self):
        """
        Processes the datetime format hints for RedshiftCopyOptions.

        NOTES:
        After testing, Redshift's date/time parsing doesn't actually
        support timezone parsing if you give it configuration - as
        documented below, it doesn't accept a time zone as part of
        the format string, and in experimentation, it silently drops
        the offset when data into a timestamptz field if you specify
        one directly.

        Redshift doesn't support time-only fields, so these will
        come in as strings regardless.

        Its automatic parser seems to be smarter, though, and is
        likely to handle a variety of formats:

        https://docs.aws.amazon.com/redshift/latest/dg/automatic-recognition.html

        The Redshift auto parser seems to take a good handling
        at our various supported formats, so let's give it a
        shot if we're not able to specify a specific format due
        to the Redshift timestamptz limitation:

        analytics=> create table formattest (test char(32));
        CREATE TABLE
        analytics=> insert into formattest values('2018-01-01 12:34:56');
        INSERT 0 1
        analytics=> insert into formattest values('01/02/18 15:34');
        INSERT 0 1
        analytics=> insert into formattest values('2018-01-02 15:34:12');
        INSERT 0 1
        analytics=> insert into formattest values('2018-01-02 10:34 PM');
        INSERT 0 1
        analytics=> select test, cast(test as timestamp) as timestamp,
                    cast(test as date) as date from formattest;

                        test               |      timestamp      |    date
        ----------------------------------+---------------------+------------
            2018-01-01 12:34:56              | 2018-01-01 12:34:56 | 2018-01-01
            01/02/18 15:34                   | 2018-01-02 15:34:00 | 2018-01-02
            2018-01-02 15:34:12              | 2018-01-02 15:34:12 | 2018-01-02
            2018-01-02 10:34 PM              | 2018-01-02 22:34:00 | 2018-01-02
        (4 rows)

        analytics=>

        """
        if self.hints.datetimeformat is None:
            self.redshift_options['time_format'] = 'auto'
        else:

            if self.hints.datetimeformat != (self
                                             .hints
                                             .datetimeformattz):
                self.redshift_options['time_format'] = 'auto'
            else:
                self.redshift_options['time_format'] = (self
                                                        .hints
                                                        .datetimeformat)
        quiet_remove(self.unhandled_hints, 'datetimeformat')
        quiet_remove(self.unhandled_hints, 'datetimeformattz')

        quiet_remove(self.unhandled_hints, 'timeonlyformat')

    def process_max_failure_rows(self):
        """
        Processes the max failure rows and fail_if_row_invalid options for RedshiftCopyOptions.
        """
        if self.max_failure_rows is not None:
            self.redshift_options['max_error'] = (self
                                                  .max_failure_rows)
        elif self.fail_if_row_invalid:
            self.redshift_options['max_error'] = 0
        else:
            self.redshift_options['max_error'] = 100000

    def process_records_terminator(self):
        """
        Processes the record terminator hint for RedshiftCopyOptions.
        """
        if self.hints.record_terminator is not None and \
                self.hints.record_terminator != "\n":
            cant_handle_hint(self.fail_if_cant_handle_hint,
                             'record-terminator', self.hints)
        quiet_remove(self.unhandled_hints, 'record-terminator')

    def process_header_row(self):
        """
        Processes the header row hint for RedshiftCopyOptions.
        """
        if self.hints.header_row:
            self.redshift_options['ignore_header'] = 1
        else:
            self.redshift_options['ignore_header'] = 0
        quiet_remove(self.unhandled_hints, 'header-row')


def redshift_copy_options(unhandled_hints: Set[str],
                          records_format: BaseRecordsFormat,
                          fail_if_cant_handle_hint: bool,
                          fail_if_row_invalid: bool,
                          max_failure_rows: Optional[int]) -> RedshiftCopyOptions:
    """
    Returns RedshiftCopyOptions based on the given parameters.

    :param unhandled_hints: Set of unhandled hints.
    :param records_format: Records format object.
    :param fail_if_cant_handle_hint: If True, raise an error if a hint is not handled.
    :param fail_if_row_invalid: If True, fail if a row is invalid.
    :param max_failure_rows: Maximum number of allowed failure rows.
    :return: A dictionary of RedshiftCopyOptions.
    """
    if isinstance(records_format, AvroRecordsFormat):
        redshift_options: RedshiftCopyOptions = dict()
        redshift_options['format'] = Format.avro
        return redshift_options
    elif isinstance(records_format, DelimitedRecordsFormat):
        redshift_delimited_records_handler = RedshiftDelimitedRecordsHandler(
            records_format, fail_if_cant_handle_hint, unhandled_hints, fail_if_row_invalid,
            max_failure_rows)
    else:
        raise NotImplementedError(f"Teach me how to COPY to {records_format}")

    return redshift_delimited_records_handler.redshift_options



import argparse
from records_mover.utils.structures import nest_dict
import os
import json
from ..mover_types import JsonSchema, JobConfig
from typing import Iterable, Dict, Any, Sequence, List, TypeVar, TYPE_CHECKING
if TYPE_CHECKING:
    from argparse_types import ArgParseArgument

ARG_VALUE_WAS_NONE: object = object()


A = TypeVar('A')


def arguments_output_to_config(arguments_output: Dict[str, A]) -> JobConfig:
    """
    argparse represents unspecified arguents as None, but we want to be
    able to know the difference between 'user told us None' (key set
    in dictionary and set to None) and 'not specified' (key not set in
    dict).
    """
    clean_config = {
        k: None if v == ARG_VALUE_WAS_NONE else v
        for k, v in arguments_output.items()
        if v is not None
    }
    return nest_dict(clean_config)


class JobConfigSchemaAsArgsParser():

    @staticmethod
    def from_description(config_json_schema: JsonSchema,
                         description: str) -> 'JobConfigSchemaAsArgsParser':
        argument_parser =\
            argparse.ArgumentParser(description=description,
                                    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        return JobConfigSchemaAsArgsParser(config_json_schema=config_json_schema,
                                           argument_parser=argument_parser)

    def __init__(self, config_json_schema: JsonSchema,
                 argument_parser: argparse.ArgumentParser) -> None:
        self.config_json_schema = config_json_schema
        self.arg_parser = argument_parser

    def add_to_prefix(self, prefix: str, key: str) -> str:
        """Return the period-separated representation of a nested JSON object.

        This allows nested JSON objects to be specified on the command
        line by representing them as a single parameter.

        (e.g., ./foo --a.b=c maps to {"a": { "b": "c"}}
        """
        if prefix == '':
            return key
        return prefix + '.' + key

    def configure_from_properties(self,
                                  properties: JsonSchema,
                                  required_keys: Iterable[str],
                                  prefix: str = '') -> None:
        self.required_keys = required_keys
        for naked_key, raw_value in properties.items():
            if not isinstance(raw_value, dict):
                raise TypeError(f"Did not understand [{raw_value}] in [{properties}]")
            value: Dict[str, object] = raw_value
            key = self.add_to_prefix(prefix, naked_key)
            self.configure_from_single_property(key, value, required_keys, properties)

    def configure_from_single_property(self,
                                       key: str,
                                       value: Dict[str, object],
                                       required_keys: Iterable[str],
                                       properties: JsonSchema) -> None:
        kwargs: ArgParseArgument = {}
        arg_name = self.get_arg_name(key, value)

        if not self.is_key_required(key):
            kwargs['required'] = False
            # When the formatted_key_name differs from key, this will make sure
            # that the arg parser is storing the value in a place that matches
            # the actual schema
            kwargs['dest'] = key

        if 'default' in value:
            kwargs['default'] = value['default']

        if 'description' in value:
            kwargs['help'] = str(value['description'])

        if 'enum' in value:
            self.add_enum_arg(key, arg_name, value, kwargs)
        elif 'type' in value and value['type'] == 'string':
            kwargs['type'] = str
            self.arg_parser.add_argument(arg_name, **kwargs)
        elif 'type' in value and value['type'] == 'integer':
            kwargs['type'] = int
            self.arg_parser.add_argument(arg_name, **kwargs)
        elif 'type' in value and value['type'] == 'array':
            self.add_array_arg(key, arg_name, value, kwargs, properties)
        elif 'type' in value and value['type'] == 'object':
            self.add_object_arg(key, arg_name, value, kwargs, properties)
        elif 'type' in value and value['type'] == 'boolean':
            self.add_bool_arg(key, arg_name, value, kwargs)
        else:
            raise Exception("Did not know how to handle key " + key +
                            " and value " + str(value))

    # For consistency with other args, we map dashes to underscores in the
    # CLI args
    def formatted_key_name(self, key: str) -> str:
        return key.replace('-', '_')

    def is_key_required(self, key: str) -> bool:
        return key in self.required_keys

    def get_arg_name(self, key: str, value: Dict[str, object]) -> str:
        if self.is_key_required(key):
            return key  # why is this not formatted_key_name?

        if value.get('type') == 'boolean' and value.get('default') is True:
            return '--no_' + self.formatted_key_name(key)
        else:
            return "--" + self.formatted_key_name(key)

    def add_enum_arg(self, key, arg_name, value, kwargs):
        enum_values: Iterable[Any] = value['enum']  # type: ignore
        non_none_values: Iterable[Any] = [v for v in enum_values if v is not None]
        kwargs['choices'] = non_none_values
        if None in enum_values:
            # add an arg that maps to 'None' for this key
            no_arg_kwargs: ArgParseArgument = {}
            # if --no_foo is specified, set foo variable to None
            split_key_name = self.formatted_key_name(key).split('.')
            no_arg_arg_name = '--' + '.'.join([*split_key_name[:-1],
                                               'no_' + split_key_name[-1]])
            no_arg_kwargs['action'] = 'store_const'
            no_arg_kwargs['const'] = ARG_VALUE_WAS_NONE
            no_arg_kwargs['dest'] = key
            no_arg_kwargs['required'] = False
            self.arg_parser.add_argument(no_arg_arg_name, **no_arg_kwargs)
        self.arg_parser.add_argument(arg_name, **kwargs)

    def add_array_arg(self, key, arg_name, value, kwargs, props):
        items = value['items']
        if not isinstance(items, dict):
            raise TypeError(f"Did not understand [{items}] in [{props}]")
        item_type = items['type']
        if not isinstance(item_type, str):
            raise TypeError(f"Did not understand [{item_type}] in [{props}]")
        if item_type == 'string':
            kwargs['type'] = str
        elif item_type == 'integer':
            kwargs['type'] = int
        else:
            raise NotImplementedError("Teach me how to handle array item "
                                      f"type of {item_type}")
        kwargs['nargs'] = '*'
        self.arg_parser.add_argument(arg_name, **kwargs)

    def add_object_arg(self, key, arg_name, value, kwargs, props):
        # Recurse and handle nested dicts
        sub_properties = value['properties']
        if not isinstance(sub_properties, dict):
            raise TypeError(f"Did not understand [{sub_properties}] in [{props}]")
        required_subkeys: List[str] = []
        if self.is_key_required(key):
            required_subkeys = [
                self.add_to_prefix(key, partial_subkey)
                for partial_subkey
                in value.get('required', [])
            ]
        sub_config = JobConfigSchemaAsArgsParser(self.config_json_schema, self.arg_parser)
        sub_config.configure_from_properties(sub_properties, required_subkeys, prefix=key)

    def add_bool_arg(self, key, arg_name, value, kwargs):
        # https://stackoverflow.com/questions/9183936/boolean-argument-for-script
        if 'default' in value:
            if value['default'] is True:
                # if --no_foo is specified, set foo variable to False
                kwargs['action'] = 'store_false'
            else:
                kwargs['action'] = 'store_true'
        else:
            if self.is_key_required(key):
                raise NotImplementedError("Teach me how to handle a required boolean "
                                          "with no default")
            else:
                # the regular --foo will set this to True:
                kwargs['action'] = 'store_const'
                # store_true sets a default, so we use store_const
                kwargs['const'] = True

                # and we'll add an arg that maps to False for this key:
                false_arg_kwargs: ArgParseArgument = {}
                # if --no_foo is specified, set foo variable to False
                split_key_name = self.formatted_key_name(key).split('.')
                false_arg_arg_name = '--' + '.'.join([*split_key_name[:-1],
                                                      'no_' + split_key_name[-1]])
                false_arg_kwargs['action'] = 'store_const'
                false_arg_kwargs['const'] = False
                false_arg_kwargs['dest'] = key
                false_arg_kwargs['required'] = False
                self.arg_parser.add_argument(false_arg_arg_name, **false_arg_kwargs)
        self.arg_parser.add_argument(arg_name, **kwargs)

    def configure_arg_parser(self) -> None:
        schema = self.config_json_schema
        if schema['type'] == 'object':
            props = schema['properties']
            if not isinstance(props, dict):
                raise TypeError(f"Did not understand {props} in {schema}")
            required_keys_raw = schema.get('required', [])
            if not isinstance(required_keys_raw, list):
                raise TypeError(f"Did not understand {required_keys_raw} in {schema}")
            required_keys: List = required_keys_raw
            self.configure_from_properties(props, required_keys)
        else:
            raise Exception("Did not know how to parse " +
                            str(self.config_json_schema))

    def parse(self, args: Sequence[str]) -> JobConfig:
        if 'JOB_CONFIG_JSON' in os.environ:
            # https://stackoverflow.com/questions/20199126/reading-json-from-a-file
            with open(os.environ['JOB_CONFIG_JSON']) as json_data:
                return json.load(json_data)
        else:
            if self.config_json_schema is None:
                return {}

            self.configure_arg_parser()
            raw_config = vars(self.arg_parser.parse_args(args))
            return arguments_output_to_config(raw_config)



from .processing_instructions import ProcessingInstructions
from .records_format import BaseRecordsFormat, DelimitedRecordsFormat


class RecordsUnloadPlan:
    def __init__(self,
                 records_format: BaseRecordsFormat = DelimitedRecordsFormat(),
                 processing_instructions: ProcessingInstructions = ProcessingInstructions()) -> \
            None:
        self.records_format = records_format
        self.processing_instructions = processing_instructions



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Optional, Set
from .types import DateOrderStyle


def determine_input_date_order_style(unhandled_hints: Set[str],
                                     hints: ValidatedRecordsHints,
                                     fail_if_cant_handle_hint: bool) ->\
        Optional[DateOrderStyle]:

    # https://www.postgresql.org/docs/9.5/datatype-datetime.html#DATATYPE-DATETIME-INPUT

    # "Date and time input is accepted in almost any reasonable
    # format, including ISO 8601, SQL-compatible, traditional
    # POSTGRES, and others. For some formats, ordering of day, month,
    # and year in date input is ambiguous and there is support for
    # specifying the expected ordering of these fields. Set the
    # DateStyle parameter to MDY to select month-day-year
    # interpretation, DMY to select day-month-year interpretation, or
    # YMD to select year-month-day interpretation."

    # $ cd tests/integration
    # $ ./itest shell
    # $ db dockerized-postgres
    # postgres=# SHOW DateStyle;
    #  DateStyle
    # -----------
    #  ISO, MDY
    # (1 row)
    #
    # postgres=#

    handler = DateInputStyleHandler(
        unhandled_hints, hints, fail_if_cant_handle_hint)

    return handler.date_order_style


class DateInputStyleHandler():
    def __init__(self,
                 unhandled_hints: Set[str],
                 hints: ValidatedRecordsHints,
                 fail_if_cant_handle_hint: bool):

        self.unhandled_hints = unhandled_hints
        self.hints = hints
        self.fail_if_cant_handle_hint = fail_if_cant_handle_hint
        self.date_order_style: Optional[DateOrderStyle] = None

        self.process_hint_datetimeformattz()
        self.process_hint_datetimeformat()
        self.process_hint_timeonlyformat()
        self.process_hint_dateformat()

    def upgrade_date_order_style(self, style: DateOrderStyle, hint_name: str) -> None:
        if self.date_order_style not in (None, style):
            cant_handle_hint(self.fail_if_cant_handle_hint, hint_name, self.hints)
        else:
            self.date_order_style = style
            quiet_remove(self.unhandled_hints, hint_name)

    def process_hint_datetimeformattz(self):
        datetimeformattz = self.hints.datetimeformattz

        # datetimeformattz: Valid values: "YYYY-MM-DD HH:MI:SSOF",
        # "YYYY-MM-DD HH:MI:SS", "YYYY-MM-DD HH24:MI:SSOF", "YYYY-MM-DD
        # HH24:MI:SSOF", "MM/DD/YY HH24:MI". See Redshift docs for more
        # information (note that HH: is equivalent to HH24: and that if
        # you don't provide an offset (OF), times are assumed to be in
        # UTC).

        # Default value is "YYYY-MM-DD HH:MI:SSOF".

        #
        # To get a postgres database to test these cases with:
        #
        # cd tests/integration
        # ./itest shell
        # db dockerized-postgres
        if datetimeformattz in ['YYYY-MM-DD HH:MI:SSOF',
                                'YYYY-MM-DD HH24:MI:SSOF']:
            #
            #  postgres=# select timestamptz '2020-01-01 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   2020-01-02 09:01:01+00
            #  (1 row)
            #
            #  postgres=#

            # Any DateStyle will do as this is unambiguous
            quiet_remove(self.unhandled_hints, 'datetimeformattz')
        elif datetimeformattz == 'YYYY-MM-DD HH:MI:SS':
            #
            #
            #  postgres=# select timestamptz '2020-01-01 23:01:01';
            #        timestamptz
            #  ------------------------
            #   2020-01-01 23:01:01+00
            #  (1 row)
            #
            #  postgres=#

            # Any DateStyle will do as this is unambiguous
            quiet_remove(self.unhandled_hints, 'datetimeformattz')
        elif datetimeformattz == "MM/DD/YY HH24:MI":
            # "MM/DD/YY HH24:MI"
            #
            #  postgres=# select timestamptz '01/02/2999 23:01';
            #        timestamptz
            #  ------------------------
            #   2999-01-02 23:01:00+00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'datetimeformattz')
        elif datetimeformattz == 'DD-MM-YY HH:MI:SSOF':
            #  postgres=# select timestamptz '02-01-2999 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   2999-02-02 09:01:01+00
            #  (1 row)
            #
            self.upgrade_date_order_style('DMY', 'datetimeformattz')
        elif datetimeformattz == 'DD/MM/YY HH:MI:SSOF':
            #  postgres=# select timestamptz '02/01/99 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   1999-02-02 09:01:01+00
            #  (1 row)
            #
            self.upgrade_date_order_style('DMY', 'datetimeformattz')
        elif datetimeformattz == 'DD-MM-YYYY HH:MI:SSOF':
            #  postgres=# select timestamptz '02-01-1999 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   1999-02-02 09:01:01+00
            #  (1 row)
            #
            self.upgrade_date_order_style('DMY', 'datetimeformattz')
        elif datetimeformattz == 'MM/DD/YY HH:MI:SSOF':
            #  postgres=# select timestamptz '01/02/99 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   1999-01-03 09:01:01+00
            #  (1 row)
            #
            self.upgrade_date_order_style('MDY', 'datetimeformattz')
        elif datetimeformattz == 'MM-DD-YYYY HH:MI:SSOF':
            #  postgres=# select timestamptz '01-02-1999 23:01:01-10';
            #        timestamptz
            #  ------------------------
            #   1999-01-03 09:01:01+00
            #  (1 row)
            #
            self.upgrade_date_order_style('MDY', 'datetimeformattz')
        else:
            cant_handle_hint(self.fail_if_cant_handle_hint,
                             'datetimeformattz', self.hints)

    def process_hint_datetimeformat(self):
        # datetimeformat: Valid values: "YYYY-MM-DD HH24:MI:SS",
        # "YYYY-MM-DD HH12:MI AM", "MM/DD/YY HH24:MI". See Redshift docs
        # for more information.

        datetimeformat = self.hints.datetimeformat

        if datetimeformat in ("YYYY-MM-DD HH24:MI:SS",
                              "YYYY-MM-DD HH:MI:SS"):
            #
            #  postgres=# select timestamp '2020-01-02 15:13:12';
            #        timestamp
            #  ---------------------
            #   2020-01-02 15:13:12
            #  (1 row)
            #
            #  postgres=#

            # Any DateStyle will do as this is unambiguous
            quiet_remove(self.unhandled_hints, 'datetimeformat')
        elif datetimeformat == "YYYY-MM-DD HH12:MI AM":
            # "YYYY-MM-DD HH12:MI AM"
            #
            #  postgres=# select timestamp '2020-01-02 1:13 PM';
            #        timestamp
            #  ---------------------
            #   2020-01-02 13:13:00
            #  (1 row)
            #
            #  postgres=#

            # Any DateStyle will do as this is unambiguous
            quiet_remove(self.unhandled_hints, 'datetimeformat')
        elif datetimeformat == "MM/DD/YY HH24:MI":
            #  postgres=# select timestamp '01/02/20 15:23';
            #        timestamp
            #  ---------------------
            #   2020-01-02 15:23:00
            #  (1 row)
            #
            #  postgres=#

            self.upgrade_date_order_style('MDY', 'datetimeformat')
        elif datetimeformat == 'DD-MM-YY HH12:MI AM':
            #  postgres=# select timestamp '02-01-20 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-02-01 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'datetimeformat')
        elif datetimeformat == 'DD/MM/YY HH12:MI AM':
            #  postgres=# select timestamp '02/01/20 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-02-01 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'datetimeformat')
        elif datetimeformat == 'DD-MM-YYYY HH12:MI AM':
            #  postgres=# select timestamp '02-01-2020 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-02-01 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'datetimeformat')
        elif datetimeformat == 'MM-DD-YY HH12:MI AM':
            #  postgres=# select timestamp '02-01-20 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-02-01 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'datetimeformat')
        elif datetimeformat == 'MM/DD/YY HH12:MI AM':
            #  postgres=# select timestamp '02/01/20 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-02-01 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'datetimeformat')
        elif datetimeformat == 'MM-DD-YYYY HH12:MI AM':
            #  postgres=# select timestamp '01-02-2020 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-01-02 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'datetimeformat')
        elif datetimeformat == 'MM-DD-YYYY HH:MI:SS':
            #  postgres=# select timestamp '01-02-2020 3:23 PM';
            #        timestamp
            #  ---------------------
            #   2020-01-02 15:23:00
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'datetimeformat')
        else:
            cant_handle_hint(self.fail_if_cant_handle_hint,
                             'datetimeformat', self.hints)

    def process_hint_timeonlyformat(self):
        timeonlyformat = self.hints.timeonlyformat

        # timeonlyformat: Valid values: "HH12:MI AM" (e.g., "1:00 PM"),
        # "HH24:MI:SS" (e.g., "13:00:00")

        if timeonlyformat == "HH12:MI AM":
            # "HH12:MI AM" (e.g., "1:00 PM"),
            #
            #  postgres=# select time '1:00 PM';
            #     time
            #  ----------
            #   13:00:00
            #  (1 row)
            #
            #  postgres=#

            # Supported!
            quiet_remove(self.unhandled_hints, 'timeonlyformat')
        elif timeonlyformat == "HH24:MI:SS":

            # "HH24:MI:SS" (e.g., "13:00:00")
            #
            #  postgres=# select time '13:00:00';
            #     time
            #  ----------
            #   13:00:00
            #  (1 row)
            #
            #  postgres=#

            # Supported!
            quiet_remove(self.unhandled_hints, 'timeonlyformat')
        elif timeonlyformat == "HH:MI:SS":
            #  postgres=# select time '13:00:00';
            #     time
            #  ----------
            #   13:00:00
            #  (1 row)
            #
            #  postgres=#
            quiet_remove(self.unhandled_hints, 'timeonlyformat')
        elif timeonlyformat == "HH24:MI":

            # "HH24:MI" (e.g., "13:00")
            #
            #  postgres=# select time '13:00';
            #     time
            #  ----------
            #   13:00:00
            #  (1 row)
            #
            #  postgres=#

            # Supported!
            quiet_remove(self.unhandled_hints, 'timeonlyformat')
        else:
            cant_handle_hint(self.fail_if_cant_handle_hint, 'timeonlyformat', self.hints)

    def process_hint_dateformat(self):
        # dateformat: Valid values:
        # null, "YYYY-MM-DD", "MM-DD-YYYY", "DD-MM-YYYY", "MM/DD/YY".
        dateformat = self.hints.dateformat

        if dateformat == "YYYY-MM-DD":
            #  postgres=# select date '1999-01-02';
            #      date
            #  ------------
            #   1999-01-02
            #  (1 row)
            #
            #  postgres=#
            # Any DateStyle will do as this is unambiguous
            quiet_remove(self.unhandled_hints, 'dateformat')
        elif dateformat == "MM-DD-YYYY":
            # "MM-DD-YYYY"
            #
            #  postgres=# select date '01-02-1999';
            #      date
            #  ------------
            #   1999-01-02
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'dateformat')
        elif dateformat == "DD-MM-YYYY":
            # "DD-MM-YYYY" - not supported by default, need to switch to
            # DMY:
            #
            #  postgres=# select date '02-01-1999';
            #      date
            #  ------------
            #   1999-02-01
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'dateformat')
        elif dateformat == "DD-MM-YY":
            #  postgres=# select date '02-01-99';
            #      date
            #  ------------
            #   1999-02-01
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'dateformat')
        elif dateformat == "MM/DD/YY":
            # "MM/DD/YY".
            #
            #  postgres=# select date '01/02/99';
            #      date
            #  ------------
            #   1999-01-02
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('MDY', 'dateformat')
        elif dateformat == "DD/MM/YY":
            #  postgres=# select date '02/01/99';
            #      date
            #  ------------
            #   1999-02-01
            #  (1 row)
            #
            #  postgres=#
            self.upgrade_date_order_style('DMY', 'dateformat')
        elif dateformat is None:
            # null implies that that date format is unknown, and that the
            # implementation SHOULD generate using their default value and
            # parse permissively.

            # ...which is what Postgres does!
            quiet_remove(self.unhandled_hints, 'dateformat')
        else:
            cant_handle_hint(self.fail_if_cant_handle_hint,
                             'dateformat', self.hints)
